﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using Opto22.Controller;

namespace DemonstrationConsole
{
	class Program
	{
		/// <summary>
		/// Update this IP address to be compatible with your PAC Controller, SoftPAC or PAC Sim.
		/// If SoftPAC or PACSim runs on the same PC as your development system, use 127.0.0.1.
		/// </summary>
		static String sControllerIpAddress = "172.22.60.3";
		/// <summary>
		/// The default TCP port used to access a controller.
		/// </summary>
		static Int32 i32ControllerPort = 22001;
		/// <summary>
		/// Initial communication timeout.
		/// </summary>
		static Int32 i32InitialTimeoutMs = 1000;

		/// <summary>
		/// Console program main.
		/// </summary>
		/// <param name="args"></param>
		static void Main(string[] args)
		{
			while (true)
			{
				// access controller properties
				MiscellaneousControllerValues();

				// access controller variables and tables via simple read methods
				ControllerVariables();

				// access controller objects using the most efficient method available
				GroupedAccessDemo();
			}
		}

		/// <summary>
		/// Demonstrate access of some controller properties.
		/// </summary>
		static void MiscellaneousControllerValues()
		{
			Int32 i32ChartControlDelayMs = 2000;

			Opto22.Controller.Controller controller = new Controller();
			Controller.ErrorResponse eMethodResponse = controller.OpenSession(sControllerIpAddress, i32ControllerPort, i32InitialTimeoutMs);
			if (eMethodResponse != Controller.ErrorResponse.Success)
			{
				Console.WriteLine("OpenSession failed with ERROR {0}.", eMethodResponse);
				return;
			}
			else
			{
				Console.WriteLine("Opening connection to {0}:{1} succeeded", sControllerIpAddress, i32ControllerPort);
			}

			Opto22.Controller.Controller.eChartStatus eStatus;
			// report the current chart status
			eMethodResponse = controller.ReadChartStatus("chtTest", false, out eStatus);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("(Prior chart status unknown) Read chart status returned value {0}.", eStatus);
			}
			else
			{
				Console.WriteLine("Read chart returned ERROR {0}.", eMethodResponse);
				return;
			}

			// change the chart state
			eMethodResponse = controller.WriteChartStatus("chtTest", Controller.Item.eItemTypes.startchart);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Chart chtTest started.");
			}
			else
			{
				Console.WriteLine("Start chart returned ERROR {0}.", eMethodResponse);
				return;
			}

			// allow chart to start
			Thread.Sleep(i32ChartControlDelayMs);

			// report the current chart status
			eMethodResponse = controller.ReadChartStatus("chtTest", false, out eStatus);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read chart status returned value {0}.", eStatus);
			}
			else
			{
				Console.WriteLine("Read chart returned ERROR {0}.", eMethodResponse);
			}


			// suspend the chart
			eMethodResponse = controller.WriteChartStatus("chtTest", Controller.Item.eItemTypes.suspendchart);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Chart chtTest suspended.");
			}
			else
			{
				Console.WriteLine("Suspend chart returned ERROR {0}.", eMethodResponse);
			}

			// allow chart to suspend
			Thread.Sleep(i32ChartControlDelayMs);

			// report the current chart status
			eMethodResponse = controller.ReadChartStatus("chtTest", false, out eStatus);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read chart status returned value {0}.", eStatus);
			}
			else
			{
				Console.WriteLine("Read chart returned ERROR {0}.", eMethodResponse);
			}

			// start the chart
			eMethodResponse = controller.WriteChartStatus("chtTest", Controller.Item.eItemTypes.continuechart);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Chart chtTest continued.");
			}
			else
			{
				Console.WriteLine("Continue chart returned ERROR {0}.", eMethodResponse);
			}

			// allow chart to continue
			Thread.Sleep(i32ChartControlDelayMs);

			// report the current chart status
			eMethodResponse = controller.ReadChartStatus("chtTest", false, out eStatus);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read chart status returned value {0}.", eStatus);
			}
			else
			{
				Console.WriteLine("Read chart returned ERROR {0}.", eMethodResponse);
			}

			// stop the chart
			eMethodResponse = controller.WriteChartStatus("chtTest", Controller.Item.eItemTypes.stopchart);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Chart chtTest stopped.");
			}
			else
			{
				Console.WriteLine("Stop chart returned ERROR {0}.", eMethodResponse);
			}

			// allow chart to stop
			Thread.Sleep(i32ChartControlDelayMs);

			// report the current chart status
			eMethodResponse = controller.ReadChartStatus("chtTest", false, out eStatus);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read chart status returned value {0}.", eStatus);
			}
			else
			{
				Console.WriteLine("Read chart returned ERROR {0}.", eMethodResponse);
			}

			// test integer "out" based reads
			Int32 i32AutoRunStatus = 0;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.autorunstatus, out i32AutoRunStatus);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read autorunstatus returned {0}.", i32AutoRunStatus);
			}
			else
			{
				Console.WriteLine("Read autorunstatus returned ERROR {0}.", eMethodResponse);
			}

			Int32 i32ChartsRunning = 0;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.chartsrunningcount, out i32ChartsRunning);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read chartsrunningcount returned {0}.", i32ChartsRunning);
			}
			else
			{
				Console.WriteLine("Read chartsrunningcount returned ERROR {0}.", eMethodResponse);
			}

			Int32 i32FileMemory = 0;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.filememory, out i32FileMemory);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read filememory returned {0}.", i32FileMemory);
			}
			else
			{
				Console.WriteLine("Read filememory returned ERROR {0}.", eMethodResponse);
			}

			String sFirmwareVersion;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.firmwareversion, out sFirmwareVersion);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read firmwareversion returned {0}.", sFirmwareVersion);
			}
			else
			{
				Console.WriteLine("Read firmwareversion returned ERROR {0}.", eMethodResponse);
			}

			Int32 i32MessageQueueCount = 0;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.getmessagequeuecount, out i32MessageQueueCount);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read getmessagequeuecount returned {0}.", i32MessageQueueCount);
			}
			else
			{
				Console.WriteLine("Read getmessagequeuecount returned ERROR {0}.", eMethodResponse);
			}

			Int32 i32StackDepth = 0;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.stackdepth, out i32StackDepth);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read stackdepth returned {0}.", i32StackDepth);
			}
			else
			{
				Console.WriteLine("Read stackdepth returned ERROR {0}.", eMethodResponse);
			}

			Int32 i32NonVolatileMemory = 0;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.nonvolatilememory, out i32NonVolatileMemory);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read nonvolatilememory returned {0}.", i32NonVolatileMemory);
			}
			else
			{
				Console.WriteLine("Read nonvolatilememory returned ERROR {0}.", eMethodResponse);
			}

			Int32 i32VolatileMemory = 0;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.volatilememory, out i32VolatileMemory);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read volatilememory returned {0}.", i32VolatileMemory);
			}
			else
			{
				Console.WriteLine("Read volatilememory returned ERROR {0}.", eMethodResponse);
			}

			Int32 i32UpTimeSeconds = 0;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.uptimeseconds, out i32UpTimeSeconds);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read uptimeseconds returned {0}.", i32UpTimeSeconds);
			}
			else
			{
				Console.WriteLine("Read uptimeseconds returned ERROR {0}.", eMethodResponse);
			}

			// test string "out" based item reads
			String sPartNumber;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.partnumber, out sPartNumber);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read partnumber returned {0}.", sPartNumber);
			}
			else
			{
				Console.WriteLine("Read partnumber returned ERROR {0}.", eMethodResponse);
			}

			String sStrategyDateStamp;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.strategydatestamp, out sStrategyDateStamp);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read strategydatestamp returned {0}.", sStrategyDateStamp);
			}
			else
			{
				Console.WriteLine("Read strategydatestamp returned ERROR {0}.", eMethodResponse);
			}

			String sStrategyName;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.strategyname, out sStrategyName);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read strategyname returned {0}.", sStrategyName);
			}
			else
			{
				Console.WriteLine("Read strategyname returned ERROR {0}.", eMethodResponse);
			}

			String sStrategyTimeStamp;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.strategytimestamp, out sStrategyTimeStamp);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read strategytimestamp returned {0}.", sStrategyTimeStamp);
			}
			else
			{
				Console.WriteLine("Read strategytimestamp returned ERROR {0}.", eMethodResponse);
			}

			String sTimeDate;
			eMethodResponse = controller.ReadItem(Controller.Item.eItemTypes.timedate, out sTimeDate);
			if (eMethodResponse == Controller.ErrorResponse.Success)
			{
				Console.WriteLine("Read timedate returned {0}.", sTimeDate);
			}
			else
			{
				Console.WriteLine("Read timedate returned ERROR {0}.", eMethodResponse);
			}

			Console.WriteLine("MiscellaneousControllerValues() finished.");
			controller.CloseSession();
			controller.Dispose();
			controller = null;
		}

		/// <summary>
		/// A simple example that just accesses the controller's variables and tables.  
		/// </summary>
		static void ControllerVariables()
		{
			// controller definition
			Opto22.Controller.Controller controller = new Opto22.Controller.Controller();

			// code generated test values to write to the controller
			Int32 i32TestValue = Int32.MinValue;
			Int64 i64TestValue = Int64.MinValue;
			Single f32TestValue = 0.0F;
			String sValue;
			Int32[] i32aryTestValues = new Int32[100];
			Int64[] i64aryTestValues = new Int64[100];
			Single[] f32aryTestValues = new Single[100];
			String[] saryTestValues = new String[100];

			// objects read from the PAC controller
			Int32 i32;
			Int64 i64;
			Single f32;
			Single f32Ut;
			Single f32Dt;
			String s;
			Int32[] i32ary;
			Int64[] i64ary;
			Single[] f32ary;
			String[] sary;

			// open a session to the controller
			Controller.ErrorResponse eMethodResponse;
			eMethodResponse = controller.OpenSession(sControllerIpAddress, i32ControllerPort, i32InitialTimeoutMs);
			// validate the attempt
			if (eMethodResponse != Controller.ErrorResponse.Success)
			{
				Console.WriteLine("OpenSession() Failed with ERROR {0}.", eMethodResponse);
				return;
			}
			else
			{
				Console.WriteLine("Opening session to {0}:{1} succeeded.", sControllerIpAddress, i32ControllerPort);
			}

			// loop to repeat reads and writes
			for (Int32 i32LoopCounter = 0; i32LoopCounter < 100; i32LoopCounter++)
			{
				#region Create Test Values To Write To Controller
				i32TestValue++;
				i64TestValue++;
				f32TestValue += 1.0F;
				sValue = i64TestValue.ToString();
				for (Int32 i = 0; i < i32aryTestValues.Length; i++) i32aryTestValues[i] = i32TestValue + i;
				for (Int32 i = 0; i < i64aryTestValues.Length; i++) i64aryTestValues[i] = i64TestValue + i;
				for (Int32 i = 0; i < f32aryTestValues.Length; i++) f32aryTestValues[i] = f32TestValue + (i * 10);
				for (Int32 i = 0; i < saryTestValues.Length; i++) saryTestValues[i] = (f32TestValue + (i * 10)).ToString();
				#endregion Create Test Values To Write To Controller

				#region Write Data To Controller
				// write variables
				eMethodResponse = controller.Write32BitIntegerVariable("i32", false, i32TestValue);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Write i32 returned {0}.", eMethodResponse);
					return;
				}

				eMethodResponse = controller.Write32BitFloatVariable("f32", false, f32TestValue);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Write f32 returned {0}.", eMethodResponse);
					return;
				}

				eMethodResponse = controller.Write64BitIntegerVariable("i64", false, i64TestValue);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Write i64 returned {0}.", eMethodResponse);
					return;
				}

				eMethodResponse = controller.WriteStringVariable("s", false, sValue);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Write s returned {0}.", eMethodResponse);
					return;
				}

				// write tables
				eMethodResponse = controller.Write32BitIntegerTable("i32t", false, i32aryTestValues, 0, i32aryTestValues.Length);
				if (eMethodResponse != Controller.ErrorResponse.Success)
					Console.WriteLine("Write i32t returned {0}.", eMethodResponse);
				eMethodResponse = controller.Write32BitFloatTable("f32t", false, f32aryTestValues, 0, f32aryTestValues.Length);
				if (eMethodResponse != Controller.ErrorResponse.Success)
					Console.WriteLine("Write f32t returned {0}.", eMethodResponse);
				eMethodResponse = controller.Write64BitIntegerTable("i64t", false, i64aryTestValues, 0, i64aryTestValues.Length);
				if (eMethodResponse != Controller.ErrorResponse.Success)
					Console.WriteLine("Write i64t returned {0}.", eMethodResponse);
				eMethodResponse = controller.WriteStringTable("st", false, saryTestValues, 0, saryTestValues.Length);
				if (eMethodResponse != Controller.ErrorResponse.Success)
					Console.WriteLine("Write st returned {0}.", eMethodResponse);
				#endregion Write Data To Controller

				#region read back test values
				eMethodResponse = controller.Read32BitIntegerVariable("i32", false, out i32);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read i32 returned {0}.", eMethodResponse);
					return;
				}
				eMethodResponse = controller.Read64BitIntegerVariable("i64", false, out i64);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read i64 returned {0}.", eMethodResponse);
					return;
				}
				eMethodResponse = controller.Read32BitFloatVariable("f32", false, out f32);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read f32 returned {0}.", eMethodResponse);
					return;
				}
				eMethodResponse = controller.ReadUpTimerVariable("ut", false, out f32Ut);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read ut returned {0}.", eMethodResponse);
					return;
				}
				eMethodResponse = controller.ReadDownTimerVariable("dt", false, out f32Dt);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read dt returned {0}.", eMethodResponse);
					return;
				}
				eMethodResponse = controller.ReadStringVariable("s", false, out s);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read s returned {0}.", eMethodResponse);
					return;
				}

				eMethodResponse = controller.Read32BitIntegerTable("i32t", false, 0, i32aryTestValues.Length, out i32ary);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read i32t returned {0}.", eMethodResponse);
					return;
				}
				eMethodResponse = controller.Read64BitIntegerTable("i64t", false, 0, f32aryTestValues.Length, out i64ary);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read i64t returned {0}.", eMethodResponse);
					return;
				}
				eMethodResponse = controller.Read32BitFloatTable("f32t", false, 0, i64aryTestValues.Length, out f32ary);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read f32t returned {0}.", eMethodResponse);
					return;
				}
				eMethodResponse = controller.ReadStringTable("st", false, 0, saryTestValues.Length, out sary);
				if (eMethodResponse != Controller.ErrorResponse.Success)
				{
					Console.WriteLine("Read st returned {0}.", eMethodResponse);
					return;
				}
				#endregion read back test values

				#region validate values read from controller
				//validate readbacks
				if (i32TestValue != i32)
				{
					Console.WriteLine("Mismatch Reading 32-bit Integer Variable");
					return;
				}
				if (i64TestValue != i64)
				{
					Console.WriteLine("Mismatch Reading 64-bit Integer Variable");
					return;
				}
				if (f32TestValue != f32)
				{
					Console.WriteLine("Mismatch Reading 32-bit Float Variable");
					return;
				}
				if (sValue != s)
				{
					Console.WriteLine("Mismatch Reading String Variable");
					return;
				}

				for (Int32 i = 0; i < i32aryTestValues.Length; i++)
				{
					if ((i32ary == null) || (i32aryTestValues[i] != i32ary[i]))
					{
						Console.WriteLine("Mismatch Reading 32-bit Integer Table Index {0}", i);
						return;
					}
				}
				for (Int32 i = 0; i < i64aryTestValues.Length; i++)
				{
					if ((i64ary == null) || (i64aryTestValues[i] != i64ary[i]))
					{
						Console.WriteLine("Mismatch Reading 64-bit Integer Table Index {0}", i);
						return;
					}
				}
				for (Int32 i = 0; i < f32aryTestValues.Length; i++)
				{
					if ((f32ary == null) || (f32aryTestValues[i] != f32ary[i]))
					{
						Console.WriteLine("Mismatch Reading 32-bit Float Table Index {0}", i);
						return;
					}
				}
				for (Int32 i = 0; i < saryTestValues.Length; i++)
				{
					if ((sary == null) || (saryTestValues[i] != sary[i]))
					{
						Console.WriteLine("Mismatch Reading String Table Index {0}", i);
						return;
					}
				}
				#endregion validate values read from controller

				#region reset read back values
				i32ary = null;
				i64ary = null;
				f32ary = null;
				sary = null;
				#endregion reset read back values

				// print activity when read values validate
				Console.WriteLine("ControllerVariables() finished scan {0}", i32LoopCounter);

				#region inspect the object's message log
				Console.WriteLine("controller.MessageLogCount reports {0} messages.", controller.MessageLogCount);
				while (controller.MessageLogCount > 0)
				{
					String sLoggedMessage;
					controller.PopMessageLogged(out sLoggedMessage);
					Console.WriteLine("  {0}", sLoggedMessage);
				}
				#endregion inspect the object's message log

				// throttle the loop's scan rate
				Thread.Sleep(250);
			}

			// close the connection to the controller
			controller.CloseSession();
			controller.Dispose();
			controller = null;
		}

		/// <summary>
		/// A demonstration of using the more efficient "grouped" data access method.
		/// </summary>
		static void GroupedAccessDemo()
		{
			// This extra opening is not required in real-world apps, it is used to test some internal DLL logic.
			Opto22.Controller.Controller controller = new Controller();
			Controller.ErrorResponse eMethodResponse = controller.OpenSession(sControllerIpAddress, i32ControllerPort, i32InitialTimeoutMs);
			if (eMethodResponse != Controller.ErrorResponse.Success)
			{
				Console.WriteLine("OpenSession failed.");
				return;
			}
			// End, extra opening.

			for (Int32 i32LoopCounter = 0; i32LoopCounter < 100; i32LoopCounter++)
			{
				const Int32 i32ItemLimit = 40;
				Controller.Item[] aryGroupItems = Controller.NewItemList(i32ItemLimit);
				Boolean bResult;
				Int32 i32ListIndex = 0;

				// analog point grouped read
				{
					#region Analog Point
					Controller.Item.eAnalogPointAttributes[] earyAnalogAllAttributes = new Controller.Item.eAnalogPointAttributes[2];
					earyAnalogAllAttributes[0] = Controller.Item.eAnalogPointAttributes.eu;
					earyAnalogAllAttributes[1] = Controller.Item.eAnalogPointAttributes.maximum;

					bResult = aryGroupItems[i32ListIndex].SetAnalogRead("ai", Controller.Item.eItemTypes.apoint, earyAnalogAllAttributes, earyAnalogAllAttributes.Length);
					i32ListIndex++;
					#endregion Analog Point

					#region Discrete Point
					Controller.Item.eDiscretePointAttributes[] earyDiscreteAllAttributes = new Controller.Item.eDiscretePointAttributes[3];
					earyDiscreteAllAttributes[0] = Controller.Item.eDiscretePointAttributes.state;
					earyDiscreteAllAttributes[1] = Controller.Item.eDiscretePointAttributes.onlatch;
					earyDiscreteAllAttributes[2] = Controller.Item.eDiscretePointAttributes.offlatch;

					bResult = aryGroupItems[i32ListIndex].SetDiscreteRead("di", Controller.Item.eItemTypes.dpoint, earyDiscreteAllAttributes, earyDiscreteAllAttributes.Length);
					i32ListIndex++;
					#endregion Discrete Point

					#region Ethernet PID
					Controller.Item.ePidAttributes[] earyPidAllAttributes = new Controller.Item.ePidAttributes[7];
					earyPidAllAttributes[0] = Controller.Item.ePidAttributes.automanual;
					earyPidAllAttributes[1] = Controller.Item.ePidAttributes.gain;
					earyPidAllAttributes[2] = Controller.Item.ePidAttributes.integral;
					earyPidAllAttributes[3] = Controller.Item.ePidAttributes.derivative;
					earyPidAllAttributes[4] = Controller.Item.ePidAttributes.input;
					earyPidAllAttributes[5] = Controller.Item.ePidAttributes.output;
					earyPidAllAttributes[6] = Controller.Item.ePidAttributes.scanrate;

					bResult = aryGroupItems[i32ListIndex].SetPidRead("pid", Controller.Item.eItemTypes.pid, earyPidAllAttributes, earyPidAllAttributes.Length);
					i32ListIndex++;
					#endregion Ethernet PID

					#region Add Reads For Variable Types
					bResult = aryGroupItems[i32ListIndex++].SetRead("i32", Controller.Item.eItemTypes.integer32);
					bResult = aryGroupItems[i32ListIndex++].SetRead("i64", Controller.Item.eItemTypes.integer64);
					bResult = aryGroupItems[i32ListIndex++].SetRead("f32", Controller.Item.eItemTypes.float32);
					bResult = aryGroupItems[i32ListIndex++].SetRead("ut", Controller.Item.eItemTypes.uptimer);
					bResult = aryGroupItems[i32ListIndex++].SetRead("s", Controller.Item.eItemTypes.stringvariable);
					bResult = aryGroupItems[i32ListIndex++].SetRead("i32t", Controller.Item.eItemTypes.integer32table, 0, 50);
					bResult = aryGroupItems[i32ListIndex++].SetRead("i64t", Controller.Item.eItemTypes.integer64table, 0, 10);
					bResult = aryGroupItems[i32ListIndex++].SetRead("f32t", Controller.Item.eItemTypes.float32table, 0, 5);
					bResult = aryGroupItems[i32ListIndex++].SetRead("st", Controller.Item.eItemTypes.stringtable, 0, 11);
					#endregion Add Reads For Variable Types
				}

				// access the items configured in aryGroupItems, all at once
				eMethodResponse = controller.ReadGroupItems(aryGroupItems, 0, i32ListIndex);
				i32ListIndex = 0;

				// test if ReadGroupItems suceeded.
				if (eMethodResponse == Controller.ErrorResponse.Success)
				{
					Boolean bGetDataResult;
					if (aryGroupItems[i32ListIndex].SuccessfulRead == true)
					{
						// process item
						Single f32AnalogEu;
						Single f32AnalogMax;
						bGetDataResult = aryGroupItems[i32ListIndex].GetAnalogData(Controller.Item.eAnalogPointAttributes.eu, out f32AnalogEu);
						if (bGetDataResult == true)
							Console.WriteLine("Analog Point eu {0}", f32AnalogEu);
						bGetDataResult = aryGroupItems[i32ListIndex].GetAnalogData(Controller.Item.eAnalogPointAttributes.maximum, out f32AnalogMax);
						if (bGetDataResult == true)
							Console.WriteLine("Analog Point maximum {0}", f32AnalogMax);
					}
					i32ListIndex++;

					if (aryGroupItems[i32ListIndex].SuccessfulRead == true)
					{
						Boolean bState;
						Boolean bOnLatch;
						Boolean bOffLatch;
						bGetDataResult = aryGroupItems[i32ListIndex].GetDiscreteData(Controller.Item.eDiscretePointAttributes.state, out bState);
						if (bGetDataResult == true)
							Console.WriteLine("Discrete state {0}", bState);
						bGetDataResult = aryGroupItems[i32ListIndex].GetDiscreteData(Controller.Item.eDiscretePointAttributes.onlatch, out bOnLatch);
						if (bGetDataResult == true)
							Console.WriteLine("Discrete on-latch {0}", bOnLatch);
						bGetDataResult = aryGroupItems[i32ListIndex].GetDiscreteData(Controller.Item.eDiscretePointAttributes.offlatch, out bOffLatch);
						if (bGetDataResult == true)
							Console.WriteLine("Discrete off-latch {0}", bOffLatch);
					}
					i32ListIndex++;

					if (aryGroupItems[i32ListIndex].SuccessfulRead == true)
					{
						Boolean bIsAutomatic;
						Single f32GainCoefficient;
						Single f32Integral;
						Single f32Derivative;
						Single f32Input;
						Single f32Output;
						Single f32Scanrate;

						bGetDataResult = aryGroupItems[i32ListIndex].GetPidData(Controller.Item.ePidAttributes.automanual, out bIsAutomatic);
						if (bGetDataResult == true)
							Console.WriteLine("PID in automatic mode {0}", bIsAutomatic);
						bGetDataResult = aryGroupItems[i32ListIndex].GetPidData(Controller.Item.ePidAttributes.gain, out f32GainCoefficient);
						if (bGetDataResult == true)
							Console.WriteLine("PID Gain {0}", f32GainCoefficient);
						bGetDataResult = aryGroupItems[i32ListIndex].GetPidData(Controller.Item.ePidAttributes.integral, out f32Integral);
						if (bGetDataResult == true)
							Console.WriteLine("PID Integral {0}", f32Integral);
						bGetDataResult = aryGroupItems[i32ListIndex].GetPidData(Controller.Item.ePidAttributes.derivative, out f32Derivative);
						if (bGetDataResult == true)
							Console.WriteLine("PID Derivative {0}", f32Derivative);
						bGetDataResult = aryGroupItems[i32ListIndex].GetPidData(Controller.Item.ePidAttributes.input, out f32Input);
						if (bGetDataResult == true)
							Console.WriteLine("PID Input {0}", f32Input);
						bGetDataResult = aryGroupItems[i32ListIndex].GetPidData(Controller.Item.ePidAttributes.output, out f32Output);
						if (bGetDataResult == true)
							Console.WriteLine("PID Output {0}", f32Output);
						bGetDataResult = aryGroupItems[i32ListIndex].GetPidData(Controller.Item.ePidAttributes.scanrate, out f32Scanrate);
						if (bGetDataResult == true)
							Console.WriteLine("PID Scan Rate {0}", f32Scanrate);
					}
					i32ListIndex++;

					Int32 i32;
					Int64 i64;
					Single f32;
					Single f32ut;
					String s;
					Int32[] i32ary;
					Int64[] i64ary;
					Single[] f32ary;
					String[] sary;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out i32);
					if (bGetDataResult == true)
						Console.WriteLine("i32 {0}", i32);
					i32ListIndex++;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out i64);
					if (bGetDataResult == true)
						Console.WriteLine("i64 {0}", i64);
					i32ListIndex++;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out f32);
					if (bGetDataResult == true)
						Console.WriteLine("f32 {0}", f32);
					i32ListIndex++;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out f32ut);
					if (bGetDataResult == true)
						Console.WriteLine("f32ut {0}", f32ut);
					i32ListIndex++;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out s);
					if (bGetDataResult == true)
						Console.WriteLine("s {0}", s);
					i32ListIndex++;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out i32ary);
					if (bGetDataResult == true)
					{
						Console.WriteLine("i32ary read successful");
					}
					i32ListIndex++;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out i64ary);
					if (bGetDataResult == true)
					{
						Console.WriteLine("i64ary read successful");
					}
					i32ListIndex++;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out f32ary);
					if (bGetDataResult == true)
					{
						Console.WriteLine("f32ary read successful");
					}
					i32ListIndex++;

					bGetDataResult = aryGroupItems[i32ListIndex].GetData(out sary);
					if (bGetDataResult == true)
					{
						Console.WriteLine("sary read successful");
					}
					i32ListIndex++;

					Thread.Sleep(0);
				}
				else
				{
					Console.WriteLine("ReadItems failed, ERROR result {0}, i32LoopCounter = {1}.", eMethodResponse, i32LoopCounter);
					break;
				}

				// throttle the scan rate
				Console.WriteLine("GroupedAccessDemo() finished scan {0}", i32LoopCounter);
				Thread.Sleep(250);
			}
			Console.WriteLine("GroupedAccessDemo() Finished.");
			controller.CloseSession();
			controller.Dispose();
			controller = null;
		}
	}
}
